// Copyright (C) 2024 Quickwit, Inc.
//
// Quickwit is offered under the AGPL v3.0 and as commercial software.
// For commercial licensing, contact us at hello@quickwit.io.
//
// AGPL:
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

use std::collections::{BTreeSet, HashMap};
use std::fmt;
use std::future::Future;
use std::time::Duration;

use fnv::{FnvHashMap, FnvHashSet};
use itertools::Itertools;
use quickwit_common::pretty::PrettySample;
use quickwit_common::Progress;
use quickwit_ingest::{IngesterPool, LocalShardsUpdate};
use quickwit_proto::control_plane::{
    AdviseResetShardsRequest, AdviseResetShardsResponse, ControlPlaneResult,
    GetOrCreateOpenShardsFailure, GetOrCreateOpenShardsFailureReason, GetOrCreateOpenShardsRequest,
    GetOrCreateOpenShardsResponse, GetOrCreateOpenShardsSuccess,
};
use quickwit_proto::ingest::ingester::{
    CloseShardsRequest, IngesterService, InitShardsRequest, RetainShardsForSource,
    RetainShardsRequest,
};
use quickwit_proto::ingest::{IngestV2Error, Shard, ShardIdPosition, ShardIdPositions, ShardIds};
use quickwit_proto::metastore;
use quickwit_proto::metastore::{MetastoreService, MetastoreServiceClient};
use quickwit_proto::types::{IndexUid, NodeId, ShardId, SourceUid};
use serde::{Deserialize, Serialize};
use tracing::{error, info, warn};
use ulid::Ulid;

use crate::ingest::wait_handle::WaitHandle;
use crate::model::{ControlPlaneModel, ScalingMode, ShardEntry, ShardStats};

const MAX_SHARD_INGESTION_THROUGHPUT_MIB_PER_SEC: f32 = 5.;

/// Threshold in MiB/s above which we increase the number of shards.
const SCALE_UP_SHARDS_THRESHOLD_MIB_PER_SEC: f32 =
    MAX_SHARD_INGESTION_THROUGHPUT_MIB_PER_SEC * 8. / 10.;

/// Threshold in MiB/s below which we decrease the number of shards.
const SCALE_DOWN_SHARDS_THRESHOLD_MIB_PER_SEC: f32 =
    MAX_SHARD_INGESTION_THROUGHPUT_MIB_PER_SEC * 2. / 10.;

const FIRE_AND_FORGET_TIMEOUT: Duration = Duration::from_secs(3);

/// Spawns a new task to execute the given future,
/// and stops polling it/drops it after a timeout.
///
/// All errors are ignored, and not even logged.
fn fire_and_forget(
    fut: impl Future<Output = ()> + Send + 'static,
    operation: impl std::fmt::Display + Send + Sync + 'static,
) {
    tokio::spawn(async move {
        if let Err(_timeout_elapsed) = tokio::time::timeout(FIRE_AND_FORGET_TIMEOUT, fut).await {
            error!(operation=%operation, "timeout elapsed");
        }
    });
}

#[derive(Debug, Default, Clone, Copy, Serialize, Deserialize)]
pub struct IngestControllerStats {
    pub num_rebalance_shards_ops: usize,
}

pub struct IngestController {
    ingester_pool: IngesterPool,
    metastore: MetastoreServiceClient,
    replication_factor: usize,
    pub stats: IngestControllerStats,
}

impl fmt::Debug for IngestController {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.debug_struct("IngestController")
            .field("ingester_pool", &self.ingester_pool)
            .field("metastore", &self.metastore)
            .field("replication_factor", &self.replication_factor)
            .finish()
    }
}

impl IngestController {
    pub fn new(
        metastore: MetastoreServiceClient,
        ingester_pool: IngesterPool,
        replication_factor: usize,
    ) -> Self {
        IngestController {
            metastore,
            ingester_pool,
            replication_factor,
            stats: IngestControllerStats::default(),
        }
    }

    /// Sends a retain shard request to the given list of ingesters.
    ///
    /// If the request fails, we just log an error.
    pub(crate) fn sync_with_ingesters(
        &self,
        ingesters: &BTreeSet<NodeId>,
        model: &ControlPlaneModel,
    ) {
        for ingester in ingesters {
            self.sync_with_ingester(ingester, model);
        }
    }

    pub(crate) fn sync_with_all_ingesters(&self, model: &ControlPlaneModel) {
        let ingesters: Vec<NodeId> = self.ingester_pool.keys();
        for ingester in ingesters {
            self.sync_with_ingester(&ingester, model);
        }
    }

    /// Syncs the ingester in a fire and forget manner.
    ///
    /// The returned oneshot is just here for unit test to wait for the operation to terminate.
    fn sync_with_ingester(&self, ingester: &NodeId, model: &ControlPlaneModel) -> WaitHandle {
        info!(ingester = %ingester, "sync_with_ingester");
        let (wait_drop_guard, wait_handle) = WaitHandle::new();
        let Some(mut ingester_client) = self.ingester_pool.get(ingester) else {
            // TODO: (Maybe) We should mark the ingester as unavailable, and stop advertise its
            // shard to routers.
            warn!("failed to sync with ingester `{ingester}`: not available");
            return wait_handle;
        };
        let mut retain_shards_req = RetainShardsRequest::default();
        for (source_uid, shard_ids) in &*model.list_shards_for_node(ingester) {
            let shards_for_source = RetainShardsForSource {
                index_uid: Some(source_uid.index_uid.clone()),
                source_id: source_uid.source_id.clone(),
                shard_ids: shard_ids.iter().cloned().collect(),
            };
            retain_shards_req
                .retain_shards_for_sources
                .push(shards_for_source);
        }
        info!(ingester = %ingester, "retain shards ingester");
        let operation: String = format!("retain shards `{ingester}`");
        fire_and_forget(
            async move {
                if let Err(retain_shards_err) =
                    ingester_client.retain_shards(retain_shards_req).await
                {
                    error!(%retain_shards_err, "retain shards error");
                }
                // just a way to force moving the drop guard.
                drop(wait_drop_guard);
            },
            operation,
        );
        wait_handle
    }

    fn handle_closed_shards(&self, closed_shards: Vec<ShardIds>, model: &mut ControlPlaneModel) {
        for closed_shard in closed_shards {
            let index_uid: IndexUid = closed_shard.index_uid().clone();
            let source_id = closed_shard.source_id;

            let source_uid = SourceUid {
                index_uid,
                source_id,
            };
            let closed_shard_ids = model.close_shards(&source_uid, &closed_shard.shard_ids);

            if !closed_shard_ids.is_empty() {
                info!(
                    index_id=%source_uid.index_uid.index_id,
                    source_id=%source_uid.source_id,
                    shard_ids=?PrettySample::new(&closed_shard_ids, 5),
                    "closed {} shard(s) reported by router",
                    closed_shard_ids.len()
                );
            }
        }
    }

    pub(crate) async fn handle_local_shards_update(
        &mut self,
        local_shards_update: LocalShardsUpdate,
        model: &mut ControlPlaneModel,
        progress: &Progress,
    ) {
        let shard_stats = model.update_shards(
            &local_shards_update.source_uid,
            &local_shards_update.shard_infos,
        );
        if shard_stats.avg_ingestion_rate >= SCALE_UP_SHARDS_THRESHOLD_MIB_PER_SEC {
            self.try_scale_up_shards(local_shards_update.source_uid, shard_stats, model, progress)
                .await;
        } else if shard_stats.avg_ingestion_rate <= SCALE_DOWN_SHARDS_THRESHOLD_MIB_PER_SEC
            && shard_stats.num_open_shards > 1
        {
            self.try_scale_down_shards(
                local_shards_update.source_uid,
                shard_stats,
                model,
                progress,
            )
            .await;
        }
    }

    fn handle_unavailable_leaders(
        &self,
        unavailable_leaders: &FnvHashSet<NodeId>,
        model: &mut ControlPlaneModel,
    ) {
        let mut confirmed_unavailable_leaders = FnvHashSet::default();

        for leader_id in unavailable_leaders {
            if !self.ingester_pool.contains_key(leader_id) {
                confirmed_unavailable_leaders.insert(leader_id.clone());
            } else {
                // TODO: If a majority of ingesters consistenly reports a leader as unavailable, we
                // should probably mark it as unavailable too.
            }
        }
        if !confirmed_unavailable_leaders.is_empty() {
            model.set_shards_as_unavailable(&confirmed_unavailable_leaders);
        }
    }

    /// Finds the open shards that satisfies the [`GetOrCreateOpenShardsRequest`] request sent by an
    /// ingest router. First, the control plane checks its internal shard table to find
    /// candidates. If it does not contain any, the control plane will ask
    /// the metastore to open new shards.
    pub(crate) async fn get_or_create_open_shards(
        &mut self,
        get_open_shards_request: GetOrCreateOpenShardsRequest,
        model: &mut ControlPlaneModel,
        progress: &Progress,
    ) -> ControlPlaneResult<GetOrCreateOpenShardsResponse> {
        self.handle_closed_shards(get_open_shards_request.closed_shards, model);

        let unavailable_leaders: FnvHashSet<NodeId> = get_open_shards_request
            .unavailable_leaders
            .into_iter()
            .map(|ingester_id| ingester_id.into())
            .collect();

        self.handle_unavailable_leaders(&unavailable_leaders, model);

        let num_subrequests = get_open_shards_request.subrequests.len();
        let mut get_or_create_open_shards_successes = Vec::with_capacity(num_subrequests);
        let mut get_or_create_open_shards_failures = Vec::new();
        let mut open_shards_subrequests = Vec::new();

        for get_open_shards_subrequest in get_open_shards_request.subrequests {
            let Some(index_uid) = model.index_uid(&get_open_shards_subrequest.index_id) else {
                let get_or_create_open_shards_failure = GetOrCreateOpenShardsFailure {
                    subrequest_id: get_open_shards_subrequest.subrequest_id,
                    index_id: get_open_shards_subrequest.index_id,
                    source_id: get_open_shards_subrequest.source_id,
                    reason: GetOrCreateOpenShardsFailureReason::IndexNotFound as i32,
                };
                get_or_create_open_shards_failures.push(get_or_create_open_shards_failure);
                continue;
            };
            let Some(open_shard_entries) = model.find_open_shards(
                &index_uid,
                &get_open_shards_subrequest.source_id,
                &unavailable_leaders,
            ) else {
                let get_or_create_open_shards_failure = GetOrCreateOpenShardsFailure {
                    subrequest_id: get_open_shards_subrequest.subrequest_id,
                    index_id: get_open_shards_subrequest.index_id,
                    source_id: get_open_shards_subrequest.source_id,
                    reason: GetOrCreateOpenShardsFailureReason::SourceNotFound as i32,
                };
                get_or_create_open_shards_failures.push(get_or_create_open_shards_failure);
                continue;
            };
            if !open_shard_entries.is_empty() {
                let open_shards: Vec<Shard> = open_shard_entries
                    .into_iter()
                    .map(|shard_entry| shard_entry.shard)
                    .collect();
                let get_or_create_open_shards_success = GetOrCreateOpenShardsSuccess {
                    subrequest_id: get_open_shards_subrequest.subrequest_id,
                    index_uid: index_uid.into(),
                    source_id: get_open_shards_subrequest.source_id,
                    open_shards,
                };
                get_or_create_open_shards_successes.push(get_or_create_open_shards_success);
            } else {
                let shard_id = ShardId::from(Ulid::new());
                let open_shards_subrequest = metastore::OpenShardsSubrequest {
                    subrequest_id: get_open_shards_subrequest.subrequest_id,
                    index_uid: index_uid.into(),
                    source_id: get_open_shards_subrequest.source_id,
                    shard_id: Some(shard_id),
                    // These attributes will be overwritten in the next stage.
                    leader_id: "".to_string(),
                    follower_id: None,
                };
                open_shards_subrequests.push(open_shards_subrequest);
            }
        }
        if !open_shards_subrequests.is_empty() {
            if let Some(leader_follower_pairs) =
                self.allocate_shards(open_shards_subrequests.len(), &unavailable_leaders, model)
            {
                for (open_shards_subrequest, (leader_id, follower_opt)) in open_shards_subrequests
                    .iter_mut()
                    .zip(leader_follower_pairs)
                {
                    open_shards_subrequest.leader_id = leader_id.into();
                    open_shards_subrequest.follower_id = follower_opt.map(Into::into);
                }
                let open_shards_request = metastore::OpenShardsRequest {
                    subrequests: open_shards_subrequests,
                };
                let open_shards_response = progress
                    .protect_future(self.metastore.open_shards(open_shards_request))
                    .await?;

                // TODO: Handle failures.
                let _ = self.init_shards(&open_shards_response, progress).await;

                for open_shards_subresponse in open_shards_response.subresponses {
                    let index_uid: IndexUid = open_shards_subresponse.index_uid().clone();
                    let source_id = open_shards_subresponse.source_id.clone();
                    model.insert_shards(
                        &index_uid,
                        &source_id,
                        open_shards_subresponse.opened_shards,
                    );
                    if let Some(open_shard_entries) =
                        model.find_open_shards(&index_uid, &source_id, &unavailable_leaders)
                    {
                        let open_shards = open_shard_entries
                            .into_iter()
                            .map(|shard_entry| shard_entry.shard)
                            .collect();
                        let get_or_create_open_shards_success = GetOrCreateOpenShardsSuccess {
                            subrequest_id: open_shards_subresponse.subrequest_id,
                            index_uid: index_uid.into(),
                            source_id: open_shards_subresponse.source_id,
                            open_shards,
                        };
                        get_or_create_open_shards_successes.push(get_or_create_open_shards_success);
                    }
                }
            } else {
                for open_shards_subrequest in open_shards_subrequests {
                    let get_or_create_open_shards_failure = GetOrCreateOpenShardsFailure {
                        subrequest_id: open_shards_subrequest.subrequest_id,
                        index_id: open_shards_subrequest.index_uid().index_id.clone(),
                        source_id: open_shards_subrequest.source_id,
                        reason: GetOrCreateOpenShardsFailureReason::NoIngestersAvailable as i32,
                    };
                    get_or_create_open_shards_failures.push(get_or_create_open_shards_failure);
                }
            }
        }
        let response = GetOrCreateOpenShardsResponse {
            successes: get_or_create_open_shards_successes,
            failures: get_or_create_open_shards_failures,
        };
        Ok(response)
    }

    /// Allocates and assigns new shards to ingesters.
    fn allocate_shards(
        &self,
        num_shards_to_allocate: usize,
        unavailable_leaders: &FnvHashSet<NodeId>,
        model: &ControlPlaneModel,
    ) -> Option<Vec<(NodeId, Option<NodeId>)>> {
        let ingesters: Vec<NodeId> = self
            .ingester_pool
            .keys()
            .into_iter()
            .filter(|ingester| !unavailable_leaders.contains(ingester))
            .sorted_by(|left, right| left.cmp(right))
            .collect();

        let num_ingesters = ingesters.len();

        if num_ingesters == 0 {
            warn!("failed to allocate {num_shards_to_allocate} shards: no ingesters available");
            return None;
        } else if self.replication_factor > num_ingesters {
            warn!(
                "failed to allocate {num_shards_to_allocate} shards: replication factor is \
                 greater than the number of available ingesters"
            );
            return None;
        }
        let mut leader_follower_pairs = Vec::with_capacity(num_shards_to_allocate);

        let mut num_open_shards: usize = 0;
        let mut num_open_shards_per_ingester: HashMap<&str, usize> =
            HashMap::with_capacity(num_ingesters);

        for shard in model.all_shards() {
            if shard.is_open() && !unavailable_leaders.contains(&shard.leader_id) {
                num_open_shards += 1;

                *num_open_shards_per_ingester
                    .entry(&shard.leader_id)
                    .or_default() += 1;
            }
        }
        let mut num_remaining_shards_to_allocate = num_shards_to_allocate;
        let num_open_shards_target = num_shards_to_allocate + num_open_shards;
        let max_num_shards_to_allocate_per_node = num_open_shards_target / num_ingesters;

        // Allocate at most `max_num_shards_to_allocate_per_node` shards to each ingester.
        for (leader_id, follower_id) in ingesters.iter().zip(ingesters.iter().cycle().skip(1)) {
            if num_remaining_shards_to_allocate == 0 {
                break;
            }
            let num_open_shards_inner = num_open_shards_per_ingester
                .get(leader_id.as_str())
                .copied()
                .unwrap_or_default();

            let num_shards_to_allocate_inner = max_num_shards_to_allocate_per_node
                .saturating_sub(num_open_shards_inner)
                .min(num_remaining_shards_to_allocate);

            for _ in 0..num_shards_to_allocate_inner {
                num_remaining_shards_to_allocate -= 1;

                let leader = leader_id.clone();
                let mut follower_opt = None;

                if self.replication_factor > 1 {
                    follower_opt = Some(follower_id.clone());
                }
                leader_follower_pairs.push((leader, follower_opt));
            }
        }
        // Allocate remaining shards one by one.
        for (leader_id, follower_id) in ingesters.iter().zip(ingesters.iter().cycle().skip(1)) {
            if num_remaining_shards_to_allocate == 0 {
                break;
            }
            num_remaining_shards_to_allocate -= 1;

            let leader = leader_id.clone();
            let mut follower_opt = None;

            if self.replication_factor > 1 {
                follower_opt = Some(follower_id.clone());
            }
            leader_follower_pairs.push((leader, follower_opt));
        }
        Some(leader_follower_pairs)
    }

    /// Calls init shards on the leaders hosting newly opened shards.
    // TODO: Return partial failures instead of failing the whole request.
    async fn init_shards(
        &self,
        open_shards_response: &metastore::OpenShardsResponse,
        progress: &Progress,
    ) -> Result<(), IngestV2Error> {
        let mut per_leader_opened_shards: FnvHashMap<&String, Vec<Shard>> = FnvHashMap::default();

        for subresponse in &open_shards_response.subresponses {
            for shard in &subresponse.opened_shards {
                per_leader_opened_shards
                    .entry(&shard.leader_id)
                    .or_default()
                    .push(shard.clone());
            }
        }
        // TODO: Init shards in parallel.
        for (leader_id, shards) in per_leader_opened_shards {
            let init_shards_request = InitShardsRequest { shards };

            let Some(mut leader) = self.ingester_pool.get(leader_id) else {
                warn!("failed to init shards: ingester `{leader_id}` is unavailable");
                continue;
            };
            progress
                .protect_future(leader.init_shards(init_shards_request))
                .await?;
        }
        Ok(())
    }

    /// Attempts to increase the number of shards. This operation is rate limited to avoid creating
    /// to many shards in a short period of time. As a result, this method may not create any
    /// shard.
    async fn try_scale_up_shards(
        &mut self,
        source_uid: SourceUid,
        shard_stats: ShardStats,
        model: &mut ControlPlaneModel,
        progress: &Progress,
    ) {
        const NUM_PERMITS: u64 = 1;

        if !model
            .acquire_scaling_permits(&source_uid, ScalingMode::Up, NUM_PERMITS)
            .unwrap_or(false)
        {
            return;
        }
        let new_num_open_shards = shard_stats.num_open_shards + 1;

        info!(
            index_id=%source_uid.index_uid.index_id,
            source_id=%source_uid.source_id,
            "scaling up number of shards to {new_num_open_shards}"
        );
        let unavailable_leaders: FnvHashSet<NodeId> = FnvHashSet::default();

        let Some((leader_id, follower_id)) = self
            .allocate_shards(1, &unavailable_leaders, model)
            .and_then(|pairs| pairs.into_iter().next())
        else {
            warn!("failed to scale up number of shards: no ingesters available");
            model.release_scaling_permits(&source_uid, ScalingMode::Up, NUM_PERMITS);
            return;
        };
        let shard_id = ShardId::from(Ulid::new());
        let open_shards_subrequest = metastore::OpenShardsSubrequest {
            subrequest_id: 0,
            index_uid: source_uid.index_uid.clone().into(),
            source_id: source_uid.source_id.clone(),
            shard_id: Some(shard_id),
            leader_id: leader_id.to_string(),
            follower_id: follower_id.map(Into::into),
        };
        let open_shards_request = metastore::OpenShardsRequest {
            subrequests: vec![open_shards_subrequest],
        };
        let open_shards_response = match progress
            .protect_future(self.metastore.open_shards(open_shards_request))
            .await
        {
            Ok(open_shards_response) => open_shards_response,
            Err(error) => {
                warn!("failed to scale up number of shards: {error}");
                model.release_scaling_permits(&source_uid, ScalingMode::Up, NUM_PERMITS);
                return;
            }
        };
        if let Err(error) = self.init_shards(&open_shards_response, progress).await {
            warn!("failed to scale up number of shards: {error}");
            model.release_scaling_permits(&source_uid, ScalingMode::Up, NUM_PERMITS);
            return;
        }
        for open_shards_subresponse in open_shards_response.subresponses {
            let index_uid = open_shards_subresponse.index_uid().clone();
            let source_id = open_shards_subresponse.source_id;

            model.insert_shards(
                &index_uid,
                &source_id,
                open_shards_subresponse.opened_shards,
            );
        }
    }

    /// Attempts to decrease the number of shards. This operation is rate limited to avoid closing
    /// shards too aggressively. As a result, this method may not close any shard.
    async fn try_scale_down_shards(
        &self,
        source_uid: SourceUid,
        shard_stats: ShardStats,
        model: &mut ControlPlaneModel,
        progress: &Progress,
    ) {
        const NUM_PERMITS: u64 = 1;

        if !model
            .acquire_scaling_permits(&source_uid, ScalingMode::Down, NUM_PERMITS)
            .unwrap_or(false)
        {
            return;
        }
        let new_num_open_shards = shard_stats.num_open_shards - 1;

        info!(
            index_id=%source_uid.index_uid.index_id,
            source_id=%source_uid.source_id,
            "scaling down number of shards to {new_num_open_shards}"
        );
        let Some((leader_id, shard_id)) = find_scale_down_candidate(&source_uid, model) else {
            model.release_scaling_permits(&source_uid, ScalingMode::Down, NUM_PERMITS);
            return;
        };
        let Some(mut ingester) = self.ingester_pool.get(&leader_id) else {
            model.release_scaling_permits(&source_uid, ScalingMode::Down, NUM_PERMITS);
            return;
        };
        let shards = vec![ShardIds {
            index_uid: source_uid.index_uid.clone().into(),
            source_id: source_uid.source_id.clone(),
            shard_ids: vec![shard_id.clone()],
        }];
        let close_shards_request = CloseShardsRequest { shards };

        if let Err(error) = progress
            .protect_future(ingester.close_shards(close_shards_request))
            .await
        {
            warn!("failed to scale down number of shards: {error}");
            model.release_scaling_permits(&source_uid, ScalingMode::Down, NUM_PERMITS);
            return;
        }
        model.close_shards(&source_uid, &[shard_id]);
    }

    pub(crate) fn advise_reset_shards(
        &self,
        request: AdviseResetShardsRequest,
        model: &ControlPlaneModel,
    ) -> AdviseResetShardsResponse {
        let mut shards_to_delete: Vec<ShardIds> = Vec::new();
        let mut shards_to_truncate: Vec<ShardIdPositions> = Vec::new();

        for shard_ids in request.shard_ids {
            let index_uid = shard_ids.index_uid().clone();
            let source_id = shard_ids.source_id.clone();

            let source_uid = SourceUid {
                index_uid,
                source_id,
            };
            let Some(shard_entries) = model.get_shards_for_source(&source_uid) else {
                // The source no longer exists: we can safely delete all the shards.
                shards_to_delete.push(shard_ids);
                continue;
            };
            let mut shard_ids_to_delete = Vec::new();
            let mut shard_positions_to_truncate = Vec::new();

            for shard_id in shard_ids.shard_ids {
                if let Some(shard_entry) = shard_entries.get(&shard_id) {
                    let publish_position_inclusive =
                        shard_entry.publish_position_inclusive().clone();

                    shard_positions_to_truncate.push(ShardIdPosition {
                        shard_id: Some(shard_id),
                        publish_position_inclusive: Some(publish_position_inclusive),
                    });
                } else {
                    shard_ids_to_delete.push(shard_id);
                }
            }
            if !shard_ids_to_delete.is_empty() {
                shards_to_delete.push(ShardIds {
                    index_uid: Some(source_uid.index_uid.clone()),
                    source_id: source_uid.source_id.clone(),
                    shard_ids: shard_ids_to_delete,
                });
            }
            if !shard_positions_to_truncate.is_empty() {
                shards_to_truncate.push(ShardIdPositions {
                    index_uid: Some(source_uid.index_uid),
                    source_id: source_uid.source_id,
                    shard_positions: shard_positions_to_truncate,
                });
            }
        }
        AdviseResetShardsResponse {
            shards_to_delete,
            shards_to_truncate,
        }
    }

    pub fn rebalance_shards(&mut self) {
        // TODO: As of now, it is only used for unit testing.
        self.stats.num_rebalance_shards_ops += 1;
    }
}

/// Finds the shard with the highest ingestion rate on the ingester with the least number of open
/// shards. If multiple shards have the same ingestion rate, the shard with the highest shard ID is
/// chosen.
fn find_scale_down_candidate(
    source_uid: &SourceUid,
    model: &ControlPlaneModel,
) -> Option<(NodeId, ShardId)> {
    let mut per_leader_candidates: HashMap<&String, (usize, &ShardEntry)> = HashMap::new();

    for shard in model.get_shards_for_source(source_uid)?.values() {
        if shard.is_open() {
            per_leader_candidates
                .entry(&shard.leader_id)
                .and_modify(|(num_shards, candidate)| {
                    *num_shards += 1;

                    if shard
                        .ingestion_rate
                        .cmp(&candidate.ingestion_rate)
                        .then_with(|| shard.shard_id.cmp(&candidate.shard_id))
                        .is_gt()
                    {
                        *candidate = shard;
                    }
                })
                .or_insert((1, shard));
        }
    }
    per_leader_candidates
        .into_iter()
        .min_by_key(|(_leader_id, (num_shards, _shard))| *num_shards)
        .map(|(leader_id, (_num_shards, shard))| {
            (leader_id.clone().into(), shard.shard_id().clone())
        })
}

#[cfg(test)]
mod tests {

    use std::collections::BTreeSet;
    use std::str::FromStr;
    use std::sync::atomic::{AtomicUsize, Ordering};
    use std::sync::Arc;

    use quickwit_config::{SourceConfig, INGEST_V2_SOURCE_ID};
    use quickwit_ingest::{RateMibPerSec, ShardInfo};
    use quickwit_metastore::IndexMetadata;
    use quickwit_proto::control_plane::GetOrCreateOpenShardsSubrequest;
    use quickwit_proto::ingest::ingester::{
        CloseShardsResponse, IngesterServiceClient, InitShardsResponse, MockIngesterService,
        RetainShardsResponse,
    };
    use quickwit_proto::ingest::{Shard, ShardState};
    use quickwit_proto::metastore::MetastoreError;
    use quickwit_proto::types::{Position, SourceId};

    use super::*;

    #[tokio::test]
    async fn test_ingest_controller_get_or_create_open_shards() {
        let source_id: &'static str = "test-source";

        let index_id_0 = "test-index-0";
        let index_metadata_0 = IndexMetadata::for_test(index_id_0, "ram://indexes/test-index-0");
        let index_uid_0 = index_metadata_0.index_uid.clone();

        let index_id_1 = "test-index-1";
        let index_metadata_1 = IndexMetadata::for_test(index_id_1, "ram://indexes/test-index-1");
        let index_uid_1 = index_metadata_1.index_uid.clone();

        let progress = Progress::default();

        let mut mock_metastore = MetastoreServiceClient::mock();
        mock_metastore.expect_open_shards().once().returning({
            let index_uid_1 = index_uid_1.clone();

            move |request| {
                assert_eq!(request.subrequests.len(), 1);
                assert_eq!(request.subrequests[0].index_uid(), &index_uid_1);
                assert_eq!(&request.subrequests[0].source_id, source_id);

                let subresponses = vec![metastore::OpenShardsSubresponse {
                    subrequest_id: 1,
                    index_uid: index_uid_1.clone().into(),
                    source_id: source_id.to_string(),
                    opened_shards: vec![Shard {
                        index_uid: index_uid_1.clone().into(),
                        source_id: source_id.to_string(),
                        shard_id: Some(ShardId::from(1)),
                        shard_state: ShardState::Open as i32,
                        leader_id: "test-ingester-2".to_string(),
                        ..Default::default()
                    }],
                }];
                let response = metastore::OpenShardsResponse { subresponses };
                Ok(response)
            }
        });
        let ingester_pool = IngesterPool::default();

        let mock_ingester = MockIngesterService::default();
        let ingester: IngesterServiceClient = mock_ingester.into();
        ingester_pool.insert("test-ingester-1".into(), ingester.clone());

        let mut mock_ingester = MockIngesterService::default();
        let index_uid_1_clone = index_uid_1.clone();
        mock_ingester
            .expect_init_shards()
            .once()
            .returning(move |request| {
                assert_eq!(request.shards.len(), 1);
                assert_eq!(request.shards[0].index_uid(), &index_uid_1_clone);
                assert_eq!(request.shards[0].source_id, "test-source");
                assert_eq!(request.shards[0].shard_id(), ShardId::from(1));
                assert_eq!(request.shards[0].leader_id, "test-ingester-2");

                Ok(InitShardsResponse {})
            });
        let ingester: IngesterServiceClient = mock_ingester.into();
        ingester_pool.insert("test-ingester-2".into(), ingester.clone());

        let replication_factor = 2;
        let mut ingest_controller = IngestController::new(
            MetastoreServiceClient::from(mock_metastore),
            ingester_pool.clone(),
            replication_factor,
        );

        let mut model = ControlPlaneModel::default();
        model.add_index(index_metadata_0.clone());
        model.add_index(index_metadata_1.clone());

        let mut source_config = SourceConfig::ingest_v2();
        source_config.source_id = source_id.to_string();

        model
            .add_source(&index_uid_0, source_config.clone())
            .unwrap();
        model.add_source(&index_uid_1, source_config).unwrap();

        let shards = vec![
            Shard {
                index_uid: index_uid_0.clone().into(),
                source_id: source_id.to_string(),
                shard_id: Some(ShardId::from(1)),
                leader_id: "test-ingester-0".to_string(),
                shard_state: ShardState::Open as i32,
                ..Default::default()
            },
            Shard {
                index_uid: index_uid_0.clone().into(),
                source_id: source_id.to_string(),
                shard_id: Some(ShardId::from(2)),
                leader_id: "test-ingester-1".to_string(),
                shard_state: ShardState::Open as i32,
                ..Default::default()
            },
        ];

        model.insert_shards(&index_uid_0, &source_id.into(), shards);

        let request = GetOrCreateOpenShardsRequest {
            subrequests: Vec::new(),
            closed_shards: Vec::new(),
            unavailable_leaders: Vec::new(),
        };

        let response = ingest_controller
            .get_or_create_open_shards(request, &mut model, &progress)
            .await
            .unwrap();

        assert_eq!(response.successes.len(), 0);
        assert_eq!(response.failures.len(), 0);

        let subrequests = vec![
            GetOrCreateOpenShardsSubrequest {
                subrequest_id: 0,
                index_id: "test-index-0".to_string(),
                source_id: source_id.to_string(),
            },
            GetOrCreateOpenShardsSubrequest {
                subrequest_id: 1,
                index_id: "test-index-1".to_string(),
                source_id: source_id.to_string(),
            },
            GetOrCreateOpenShardsSubrequest {
                subrequest_id: 2,
                index_id: "index-not-found".to_string(),
                source_id: "source-not-found".to_string(),
            },
            GetOrCreateOpenShardsSubrequest {
                subrequest_id: 3,
                index_id: "test-index-0".to_string(),
                source_id: "source-not-found".to_string(),
            },
        ];
        let closed_shards = Vec::new();
        let unavailable_leaders = vec!["test-ingester-0".to_string()];
        let request = GetOrCreateOpenShardsRequest {
            subrequests,
            closed_shards,
            unavailable_leaders,
        };
        let response = ingest_controller
            .get_or_create_open_shards(request, &mut model, &progress)
            .await
            .unwrap();

        assert_eq!(response.successes.len(), 2);
        assert_eq!(response.failures.len(), 2);

        let success = &response.successes[0];
        assert_eq!(success.subrequest_id, 0);
        assert_eq!(success.index_uid(), &index_uid_0);
        assert_eq!(success.source_id, source_id);
        assert_eq!(success.open_shards.len(), 1);
        assert_eq!(success.open_shards[0].shard_id(), ShardId::from(2));
        assert_eq!(success.open_shards[0].leader_id, "test-ingester-1");

        let success = &response.successes[1];
        assert_eq!(success.subrequest_id, 1);
        assert_eq!(success.index_uid(), &index_uid_1);
        assert_eq!(success.source_id, source_id);
        assert_eq!(success.open_shards.len(), 1);
        assert_eq!(success.open_shards[0].shard_id(), ShardId::from(1));
        assert_eq!(success.open_shards[0].leader_id, "test-ingester-2");

        let failure = &response.failures[0];
        assert_eq!(failure.subrequest_id, 2);
        assert_eq!(failure.index_id, "index-not-found");
        assert_eq!(failure.source_id, "source-not-found");
        assert_eq!(
            failure.reason(),
            GetOrCreateOpenShardsFailureReason::IndexNotFound
        );

        let failure = &response.failures[1];
        assert_eq!(failure.subrequest_id, 3);
        assert_eq!(failure.index_id, index_id_0);
        assert_eq!(failure.source_id, "source-not-found");
        assert_eq!(
            failure.reason(),
            GetOrCreateOpenShardsFailureReason::SourceNotFound
        );

        assert_eq!(model.num_shards(), 3);
    }

    #[tokio::test]
    async fn test_ingest_controller_get_open_shards_handles_closed_shards() {
        let metastore = MetastoreServiceClient::mock().into();
        let ingester_pool = IngesterPool::default();
        let replication_factor = 2;

        let mut ingest_controller =
            IngestController::new(metastore, ingester_pool, replication_factor);
        let mut model = ControlPlaneModel::default();

        let index_uid = IndexUid::for_test("test-index-0", 0);
        let source_id: SourceId = "test-source".into();

        let shards = vec![Shard {
            shard_id: Some(ShardId::from(1)),
            index_uid: Some(index_uid.clone()),
            source_id: source_id.clone(),
            leader_id: "test-ingester-0".to_string(),
            shard_state: ShardState::Open as i32,
            ..Default::default()
        }];
        model.insert_shards(&index_uid, &source_id, shards);

        let request = GetOrCreateOpenShardsRequest {
            subrequests: Vec::new(),
            closed_shards: vec![ShardIds {
                index_uid: index_uid.clone().into(),
                source_id: source_id.clone(),
                shard_ids: vec![ShardId::from(1), ShardId::from(2)],
            }],
            unavailable_leaders: Vec::new(),
        };
        let progress = Progress::default();

        ingest_controller
            .get_or_create_open_shards(request, &mut model, &progress)
            .await
            .unwrap();

        let shard_1 = model
            .all_shards()
            .find(|shard| shard.shard_id() == ShardId::from(1))
            .unwrap();
        assert!(shard_1.is_closed());
    }

    #[tokio::test]
    async fn test_ingest_controller_get_open_shards_handles_unavailable_leaders() {
        let metastore = MetastoreServiceClient::mock().into();

        let ingester_pool = IngesterPool::default();
        let ingester_1 = IngesterServiceClient::mock().into();
        ingester_pool.insert("test-ingester-1".into(), ingester_1);

        let replication_factor = 2;

        let mut ingest_controller =
            IngestController::new(metastore, ingester_pool.clone(), replication_factor);
        let mut model = ControlPlaneModel::default();

        let index_uid = IndexUid::for_test("test-index-0", 0);
        let source_id: SourceId = "test-source".into();

        let shards = vec![
            Shard {
                index_uid: Some(index_uid.clone()),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(1)),
                leader_id: "test-ingester-0".to_string(),
                shard_state: ShardState::Open as i32,
                ..Default::default()
            },
            Shard {
                index_uid: Some(index_uid.clone()),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(2)),
                leader_id: "test-ingester-0".to_string(),
                shard_state: ShardState::Closed as i32,
                ..Default::default()
            },
            Shard {
                index_uid: Some(index_uid.clone()),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(3)),
                leader_id: "test-ingester-1".to_string(),
                shard_state: ShardState::Open as i32,
                ..Default::default()
            },
        ];
        model.insert_shards(&index_uid, &source_id, shards);

        let request = GetOrCreateOpenShardsRequest {
            subrequests: Vec::new(),
            closed_shards: Vec::new(),
            unavailable_leaders: vec!["test-ingester-0".to_string()],
        };
        let progress = Progress::default();

        ingest_controller
            .get_or_create_open_shards(request, &mut model, &progress)
            .await
            .unwrap();

        let shard_1 = model
            .all_shards()
            .find(|shard| shard.shard_id() == ShardId::from(1))
            .unwrap();
        assert!(shard_1.is_unavailable());

        let shard_2 = model
            .all_shards()
            .find(|shard| shard.shard_id() == ShardId::from(2))
            .unwrap();
        assert!(shard_2.is_closed());

        let shard_3 = model
            .all_shards()
            .find(|shard| shard.shard_id() == ShardId::from(3))
            .unwrap();
        assert!(shard_3.is_open());
    }

    #[test]
    fn test_ingest_controller_allocate_shards() {
        let metastore = MetastoreServiceClient::mock().into();
        let ingester_pool = IngesterPool::default();
        let replication_factor = 2;

        let ingest_controller =
            IngestController::new(metastore, ingester_pool.clone(), replication_factor);

        let mut model = ControlPlaneModel::default();

        let leader_follower_pairs_opt =
            ingest_controller.allocate_shards(0, &FnvHashSet::default(), &model);
        assert!(leader_follower_pairs_opt.is_none());

        ingester_pool.insert(
            "test-ingester-1".into(),
            IngesterServiceClient::mock().into(),
        );

        let leader_follower_pairs_opt =
            ingest_controller.allocate_shards(0, &FnvHashSet::default(), &model);
        assert!(leader_follower_pairs_opt.is_none());

        ingester_pool.insert(
            "test-ingester-2".into(),
            IngesterServiceClient::mock().into(),
        );

        let leader_follower_pairs = ingest_controller
            .allocate_shards(0, &FnvHashSet::default(), &model)
            .unwrap();
        assert!(leader_follower_pairs.is_empty());

        let leader_follower_pairs = ingest_controller
            .allocate_shards(1, &FnvHashSet::default(), &model)
            .unwrap();
        assert_eq!(leader_follower_pairs.len(), 1);
        assert_eq!(leader_follower_pairs[0].0, "test-ingester-1");
        assert_eq!(
            leader_follower_pairs[0].1,
            Some(NodeId::from("test-ingester-2"))
        );

        let leader_follower_pairs = ingest_controller
            .allocate_shards(2, &FnvHashSet::default(), &model)
            .unwrap();
        assert_eq!(leader_follower_pairs.len(), 2);
        assert_eq!(leader_follower_pairs[0].0, "test-ingester-1");
        assert_eq!(
            leader_follower_pairs[0].1,
            Some(NodeId::from("test-ingester-2"))
        );

        assert_eq!(leader_follower_pairs[1].0, "test-ingester-2");
        assert_eq!(
            leader_follower_pairs[1].1,
            Some(NodeId::from("test-ingester-1"))
        );

        let leader_follower_pairs = ingest_controller
            .allocate_shards(3, &FnvHashSet::default(), &model)
            .unwrap();
        assert_eq!(leader_follower_pairs.len(), 3);
        assert_eq!(leader_follower_pairs[0].0, "test-ingester-1");
        assert_eq!(
            leader_follower_pairs[0].1,
            Some(NodeId::from("test-ingester-2"))
        );

        assert_eq!(leader_follower_pairs[1].0, "test-ingester-2");
        assert_eq!(
            leader_follower_pairs[1].1,
            Some(NodeId::from("test-ingester-1"))
        );

        assert_eq!(leader_follower_pairs[2].0, "test-ingester-1");
        assert_eq!(
            leader_follower_pairs[2].1,
            Some(NodeId::from("test-ingester-2"))
        );

        let index_uid = IndexUid::for_test("test-index", 0);
        let source_id: SourceId = "test-source".into();
        let open_shards = vec![Shard {
            index_uid: Some(index_uid.clone()),
            source_id: source_id.clone(),
            shard_id: Some(ShardId::from(1)),
            shard_state: ShardState::Open as i32,
            leader_id: "test-ingester-1".to_string(),
            ..Default::default()
        }];
        model.insert_shards(&index_uid, &source_id, open_shards);

        let leader_follower_pairs = ingest_controller
            .allocate_shards(3, &FnvHashSet::default(), &model)
            .unwrap();
        assert_eq!(leader_follower_pairs.len(), 3);
        assert_eq!(leader_follower_pairs[0].0, "test-ingester-1");
        assert_eq!(
            leader_follower_pairs[0].1,
            Some(NodeId::from("test-ingester-2"))
        );

        assert_eq!(leader_follower_pairs[1].0, "test-ingester-2");
        assert_eq!(
            leader_follower_pairs[1].1,
            Some(NodeId::from("test-ingester-1"))
        );

        assert_eq!(leader_follower_pairs[2].0, "test-ingester-2");
        assert_eq!(
            leader_follower_pairs[2].1,
            Some(NodeId::from("test-ingester-1"))
        );

        let open_shards = vec![
            Shard {
                index_uid: Some(index_uid.clone()),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(2)),
                shard_state: ShardState::Open as i32,
                leader_id: "test-ingester-1".to_string(),
                ..Default::default()
            },
            Shard {
                index_uid: Some(index_uid.clone()),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(3)),
                shard_state: ShardState::Open as i32,
                leader_id: "test-ingester-1".to_string(),
                ..Default::default()
            },
        ];
        model.insert_shards(&index_uid, &source_id, open_shards);

        let leader_follower_pairs = ingest_controller
            .allocate_shards(1, &FnvHashSet::default(), &model)
            .unwrap();
        assert_eq!(leader_follower_pairs.len(), 1);
        assert_eq!(leader_follower_pairs[0].0, "test-ingester-2");
        assert_eq!(
            leader_follower_pairs[0].1,
            Some(NodeId::from("test-ingester-1"))
        );

        ingester_pool.insert(
            "test-ingester-3".into(),
            IngesterServiceClient::mock().into(),
        );
        let unavailable_leaders = FnvHashSet::from_iter([NodeId::from("test-ingester-2")]);
        let leader_follower_pairs = ingest_controller
            .allocate_shards(4, &unavailable_leaders, &model)
            .unwrap();
        assert_eq!(leader_follower_pairs.len(), 4);
        assert_eq!(leader_follower_pairs[0].0, "test-ingester-3");
        assert_eq!(
            leader_follower_pairs[0].1,
            Some(NodeId::from("test-ingester-1"))
        );

        assert_eq!(leader_follower_pairs[1].0, "test-ingester-3");
        assert_eq!(
            leader_follower_pairs[1].1,
            Some(NodeId::from("test-ingester-1"))
        );

        assert_eq!(leader_follower_pairs[2].0, "test-ingester-3");
        assert_eq!(
            leader_follower_pairs[2].1,
            Some(NodeId::from("test-ingester-1"))
        );

        assert_eq!(leader_follower_pairs[3].0, "test-ingester-1");
        assert_eq!(
            leader_follower_pairs[3].1,
            Some(NodeId::from("test-ingester-3"))
        );
    }

    #[tokio::test]
    async fn test_ingest_controller_handle_local_shards_update() {
        let mut mock_metastore = MetastoreServiceClient::mock();
        mock_metastore
            .expect_open_shards()
            .once()
            .returning(|request| {
                assert_eq!(request.subrequests.len(), 1);
                let subrequest = &request.subrequests[0];

                assert_eq!(subrequest.index_uid(), &IndexUid::for_test("test-index", 0));
                assert_eq!(subrequest.source_id, "test-source");
                assert_eq!(subrequest.leader_id, "test-ingester");

                Err(MetastoreError::InvalidArgument {
                    message: "failed to open shards".to_string(),
                })
            });
        let metastore = MetastoreServiceClient::from(mock_metastore);
        let ingester_pool = IngesterPool::default();
        let replication_factor = 1;

        let mut ingest_controller =
            IngestController::new(metastore, ingester_pool.clone(), replication_factor);

        let index_uid = IndexUid::for_test("test-index", 0);
        let source_id: SourceId = "test-source".into();

        let source_uid = SourceUid {
            index_uid: index_uid.clone(),
            source_id: source_id.clone(),
        };
        let mut model = ControlPlaneModel::default();
        let progress = Progress::default();

        let shards = vec![Shard {
            index_uid: Some(index_uid.clone()),
            source_id: source_id.clone(),
            shard_id: Some(ShardId::from(1)),
            leader_id: "test-ingester".to_string(),
            shard_state: ShardState::Open as i32,
            ..Default::default()
        }];
        model.insert_shards(&index_uid, &source_id, shards);
        let shard_entries: Vec<ShardEntry> = model.all_shards().cloned().collect();

        assert_eq!(shard_entries.len(), 1);
        assert_eq!(shard_entries[0].ingestion_rate, 0);

        // Test update shard ingestion rate but no scale down because num open shards is 1.
        let shard_infos = BTreeSet::from_iter([ShardInfo {
            shard_id: ShardId::from(1),
            shard_state: ShardState::Open,
            ingestion_rate: RateMibPerSec(1),
        }]);
        let local_shards_update = LocalShardsUpdate {
            leader_id: "test-ingester".into(),
            source_uid: source_uid.clone(),
            shard_infos,
        };
        ingest_controller
            .handle_local_shards_update(local_shards_update, &mut model, &progress)
            .await;

        let shard_entries: Vec<ShardEntry> = model.all_shards().cloned().collect();
        assert_eq!(shard_entries.len(), 1);
        assert_eq!(shard_entries[0].ingestion_rate, 1);

        // Test update shard ingestion rate with failing scale down.
        let shards = vec![Shard {
            index_uid: Some(index_uid.clone()),
            source_id: source_id.clone(),
            shard_id: Some(ShardId::from(2)),
            shard_state: ShardState::Open as i32,
            leader_id: "test-ingester".to_string(),
            ..Default::default()
        }];
        model.insert_shards(&index_uid, &source_id, shards);

        let shard_entries: Vec<ShardEntry> = model.all_shards().cloned().collect();
        assert_eq!(shard_entries.len(), 2);

        let mut ingester_mock = IngesterServiceClient::mock();

        let index_uid_clone = index_uid.clone();
        ingester_mock
            .expect_close_shards()
            .returning(move |request| {
                assert_eq!(request.shards.len(), 1);
                assert_eq!(request.shards[0].index_uid(), &index_uid_clone);
                assert_eq!(request.shards[0].source_id, "test-source");
                assert_eq!(request.shards[0].shard_ids, [ShardId::from(2)]);

                Err(IngestV2Error::Internal(
                    "failed to close shards".to_string(),
                ))
            });
        ingester_pool.insert("test-ingester".into(), ingester_mock.into());

        let shard_infos = BTreeSet::from_iter([
            ShardInfo {
                shard_id: ShardId::from(1),
                shard_state: ShardState::Open,
                ingestion_rate: RateMibPerSec(1),
            },
            ShardInfo {
                shard_id: ShardId::from(2),
                shard_state: ShardState::Open,
                ingestion_rate: RateMibPerSec(1),
            },
        ]);
        let local_shards_update = LocalShardsUpdate {
            leader_id: "test-ingester".into(),
            source_uid: source_uid.clone(),
            shard_infos,
        };
        ingest_controller
            .handle_local_shards_update(local_shards_update, &mut model, &progress)
            .await;

        // Test update shard ingestion rate with failing scale up.
        let shard_infos = BTreeSet::from_iter([
            ShardInfo {
                shard_id: ShardId::from(1),
                shard_state: ShardState::Open,
                ingestion_rate: RateMibPerSec(4),
            },
            ShardInfo {
                shard_id: ShardId::from(2),
                shard_state: ShardState::Open,
                ingestion_rate: RateMibPerSec(4),
            },
        ]);
        let local_shards_update = LocalShardsUpdate {
            leader_id: "test-ingester".into(),
            source_uid: source_uid.clone(),
            shard_infos,
        };
        ingest_controller
            .handle_local_shards_update(local_shards_update, &mut model, &progress)
            .await;
    }

    #[tokio::test]
    async fn test_ingest_controller_try_scale_up_shards() {
        let mut mock_metastore = MetastoreServiceClient::mock();

        let index_uid = IndexUid::from_str("test-index:00000000000000000000000000").unwrap();
        let index_uid_clone = index_uid.clone();
        mock_metastore
            .expect_open_shards()
            .once()
            .returning(move |request| {
                assert_eq!(request.subrequests.len(), 1);
                assert_eq!(request.subrequests[0].index_uid(), &index_uid_clone);
                assert_eq!(request.subrequests[0].source_id, INGEST_V2_SOURCE_ID);
                assert_eq!(request.subrequests[0].leader_id, "test-ingester");

                Err(MetastoreError::InvalidArgument {
                    message: "failed to open shards".to_string(),
                })
            });
        let index_uid_clone = index_uid.clone();
        mock_metastore
            .expect_open_shards()
            .returning(move |request| {
                assert_eq!(request.subrequests.len(), 1);
                assert_eq!(request.subrequests[0].index_uid(), &index_uid_clone);
                assert_eq!(request.subrequests[0].source_id, INGEST_V2_SOURCE_ID);
                assert_eq!(request.subrequests[0].leader_id, "test-ingester");

                let subresponses = vec![metastore::OpenShardsSubresponse {
                    subrequest_id: 0,
                    index_uid: Some(index_uid.clone()),
                    source_id: INGEST_V2_SOURCE_ID.to_string(),
                    opened_shards: vec![Shard {
                        index_uid: Some(index_uid.clone()),
                        source_id: INGEST_V2_SOURCE_ID.to_string(),
                        shard_id: Some(ShardId::from(1)),
                        leader_id: "test-ingester".to_string(),
                        shard_state: ShardState::Open as i32,
                        ..Default::default()
                    }],
                }];
                let response = metastore::OpenShardsResponse { subresponses };
                Ok(response)
            });

        let ingester_pool = IngesterPool::default();
        let replication_factor = 1;

        let mut ingest_controller = IngestController::new(
            mock_metastore.into(),
            ingester_pool.clone(),
            replication_factor,
        );

        let index_uid = IndexUid::for_test("test-index", 0);
        let source_id: SourceId = INGEST_V2_SOURCE_ID.to_string();

        let source_uid = SourceUid {
            index_uid: index_uid.clone(),
            source_id: source_id.clone(),
        };
        let shard_stats = ShardStats {
            num_open_shards: 2,
            ..Default::default()
        };
        let mut model = ControlPlaneModel::default();
        let index_metadata =
            IndexMetadata::for_test(&index_uid.index_id, "ram://indexes/test-index:0");
        model.add_index(index_metadata);

        let souce_config = SourceConfig::ingest_v2();
        model.add_source(&index_uid, souce_config).unwrap();

        let progress = Progress::default();

        // Test could not find leader.
        ingest_controller
            .try_scale_up_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;

        let mut ingester_mock = IngesterServiceClient::mock();

        let index_uid_clone = index_uid.clone();
        ingester_mock
            .expect_init_shards()
            .once()
            .returning(move |request| {
                assert_eq!(request.shards.len(), 1);
                assert_eq!(request.shards[0].index_uid(), &index_uid_clone);
                assert_eq!(request.shards[0].source_id, INGEST_V2_SOURCE_ID);
                assert_eq!(request.shards[0].shard_id(), ShardId::from(1));
                assert_eq!(request.shards[0].leader_id, "test-ingester");

                Err(IngestV2Error::Internal("failed to init shards".to_string()))
            });
        let index_uid_clone = index_uid.clone();
        ingester_mock
            .expect_init_shards()
            .returning(move |request| {
                assert_eq!(request.shards.len(), 1);
                assert_eq!(request.shards[0].index_uid(), &index_uid_clone);
                assert_eq!(request.shards[0].source_id, INGEST_V2_SOURCE_ID);
                assert_eq!(request.shards[0].shard_id(), ShardId::from(1));
                assert_eq!(request.shards[0].leader_id, "test-ingester");

                Ok(InitShardsResponse {})
            });
        ingester_pool.insert("test-ingester".into(), ingester_mock.into());

        // Test failed to open shards.
        ingest_controller
            .try_scale_up_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;
        assert_eq!(model.all_shards().count(), 0);

        // Test failed to init shards.
        ingest_controller
            .try_scale_up_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;
        assert_eq!(model.all_shards().count(), 0);

        // Test successfully opened shard.
        ingest_controller
            .try_scale_up_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;
        assert_eq!(
            model.all_shards().filter(|shard| shard.is_open()).count(),
            1
        );
    }

    #[tokio::test]
    async fn test_ingest_controller_try_scale_down_shards() {
        let metastore = MetastoreServiceClient::mock().into();
        let ingester_pool = IngesterPool::default();
        let replication_factor = 1;

        let ingest_controller =
            IngestController::new(metastore, ingester_pool.clone(), replication_factor);

        let index_uid = IndexUid::for_test("test-index", 0);
        let source_id: SourceId = "test-source".into();

        let source_uid = SourceUid {
            index_uid: index_uid.clone(),
            source_id: source_id.clone(),
        };
        let shard_stats = ShardStats {
            num_open_shards: 2,
            ..Default::default()
        };
        let mut model = ControlPlaneModel::default();
        let progress = Progress::default();

        // Test could not find a scale down candidate.
        ingest_controller
            .try_scale_down_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;

        let shards = vec![Shard {
            shard_id: Some(ShardId::from(1)),
            index_uid: Some(index_uid.clone()),
            source_id: source_id.clone(),
            leader_id: "test-ingester".to_string(),
            shard_state: ShardState::Open as i32,
            ..Default::default()
        }];
        model.insert_shards(&index_uid, &source_id, shards);

        // Test ingester is unavailable.
        ingest_controller
            .try_scale_down_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;

        let mut ingester_mock = IngesterServiceClient::mock();

        let index_uid_clone = index_uid.clone();
        ingester_mock
            .expect_close_shards()
            .once()
            .returning(move |request| {
                assert_eq!(request.shards.len(), 1);
                assert_eq!(request.shards[0].index_uid(), &index_uid_clone);
                assert_eq!(request.shards[0].source_id, "test-source");
                assert_eq!(request.shards[0].shard_ids, [ShardId::from(1)]);

                Err(IngestV2Error::Internal(
                    "failed to close shards".to_string(),
                ))
            });
        let index_uid_clone = index_uid.clone();
        ingester_mock
            .expect_close_shards()
            .once()
            .returning(move |request| {
                assert_eq!(request.shards.len(), 1);
                assert_eq!(request.shards[0].index_uid(), &index_uid_clone);
                assert_eq!(request.shards[0].source_id, "test-source");
                assert_eq!(request.shards[0].shard_ids, [ShardId::from(1)]);

                Ok(CloseShardsResponse {})
            });
        ingester_pool.insert("test-ingester".into(), ingester_mock.into());

        // Test failed to close shard.
        ingest_controller
            .try_scale_down_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;
        assert!(model.all_shards().all(|shard| shard.is_open()));

        // Test successfully closed shard.
        ingest_controller
            .try_scale_down_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;
        assert!(model.all_shards().all(|shard| shard.is_closed()));

        let shards = vec![Shard {
            shard_id: Some(ShardId::from(2)),
            index_uid: Some(index_uid.clone()),
            source_id: source_id.clone(),
            leader_id: "test-ingester".to_string(),
            shard_state: ShardState::Open as i32,
            ..Default::default()
        }];
        model.insert_shards(&index_uid, &source_id, shards);

        // Test rate limited.
        ingest_controller
            .try_scale_down_shards(source_uid.clone(), shard_stats, &mut model, &progress)
            .await;
        assert!(model.all_shards().any(|shard| shard.is_open()));
    }

    #[test]
    fn test_find_scale_down_candidate() {
        let index_uid = IndexUid::for_test("test-index", 0);
        let source_id: SourceId = "test-source".into();

        let source_uid = SourceUid {
            index_uid: index_uid.clone(),
            source_id: source_id.clone(),
        };
        let mut model = ControlPlaneModel::default();

        assert!(find_scale_down_candidate(&source_uid, &model).is_none());

        let shards = vec![
            Shard {
                index_uid: index_uid.clone().into(),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(1)),
                shard_state: ShardState::Open as i32,
                leader_id: "test-ingester-0".to_string(),
                ..Default::default()
            },
            Shard {
                index_uid: index_uid.clone().into(),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(2)),
                shard_state: ShardState::Open as i32,
                leader_id: "test-ingester-0".to_string(),
                ..Default::default()
            },
            Shard {
                index_uid: index_uid.clone().into(),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(3)),
                shard_state: ShardState::Closed as i32,
                leader_id: "test-ingester-0".to_string(),
                ..Default::default()
            },
            Shard {
                index_uid: index_uid.clone().into(),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(4)),
                shard_state: ShardState::Open as i32,
                leader_id: "test-ingester-1".to_string(),
                ..Default::default()
            },
            Shard {
                index_uid: index_uid.clone().into(),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(5)),
                shard_state: ShardState::Open as i32,
                leader_id: "test-ingester-1".to_string(),
                ..Default::default()
            },
            Shard {
                index_uid: index_uid.clone().into(),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(6)),
                shard_state: ShardState::Open as i32,
                leader_id: "test-ingester-1".to_string(),
                ..Default::default()
            },
        ];
        model.insert_shards(&index_uid, &source_id, shards);

        let shard_infos = BTreeSet::from_iter([
            ShardInfo {
                shard_id: ShardId::from(1),
                shard_state: ShardState::Open,
                ingestion_rate: quickwit_ingest::RateMibPerSec(1),
            },
            ShardInfo {
                shard_id: ShardId::from(2),
                shard_state: ShardState::Open,
                ingestion_rate: quickwit_ingest::RateMibPerSec(2),
            },
            ShardInfo {
                shard_id: ShardId::from(3),
                shard_state: ShardState::Open,
                ingestion_rate: quickwit_ingest::RateMibPerSec(3),
            },
            ShardInfo {
                shard_id: ShardId::from(4),
                shard_state: ShardState::Open,
                ingestion_rate: quickwit_ingest::RateMibPerSec(4),
            },
            ShardInfo {
                shard_id: ShardId::from(5),
                shard_state: ShardState::Open,
                ingestion_rate: quickwit_ingest::RateMibPerSec(5),
            },
            ShardInfo {
                shard_id: ShardId::from(6),
                shard_state: ShardState::Open,
                ingestion_rate: quickwit_ingest::RateMibPerSec(6),
            },
        ]);
        model.update_shards(&source_uid, &shard_infos);

        let (leader_id, shard_id) = find_scale_down_candidate(&source_uid, &model).unwrap();
        assert_eq!(leader_id, "test-ingester-0");
        assert_eq!(shard_id, ShardId::from(2));
    }

    #[tokio::test]
    async fn test_sync_with_ingesters() {
        let metastore = MetastoreServiceClient::mock().into();
        let ingester_pool = IngesterPool::default();
        let replication_factor = 2;

        let ingest_controller =
            IngestController::new(metastore, ingester_pool.clone(), replication_factor);

        let index_uid = IndexUid::for_test("test-index", 0);
        let source_id: SourceId = "test-source".into();
        let mut model = ControlPlaneModel::default();
        let shards = vec![
            Shard {
                index_uid: Some(index_uid.clone()),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(1)),
                shard_state: ShardState::Open as i32,
                leader_id: "node-1".to_string(),
                follower_id: Some("node-2".to_string()),
                ..Default::default()
            },
            Shard {
                index_uid: Some(index_uid.clone()),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(2)),
                shard_state: ShardState::Open as i32,
                leader_id: "node-2".to_string(),
                follower_id: Some("node-3".to_string()),
                ..Default::default()
            },
            Shard {
                index_uid: Some(index_uid.clone()),
                source_id: source_id.clone(),
                shard_id: Some(ShardId::from(3)),
                shard_state: ShardState::Open as i32,
                leader_id: "node-2".to_string(),
                follower_id: Some("node-1".to_string()),
                ..Default::default()
            },
        ];
        model.insert_shards(&index_uid, &source_id, shards);

        let mut ingester_mock1 = IngesterServiceClient::mock();
        let ingester_mock2 = IngesterServiceClient::mock();
        let ingester_mock3 = IngesterServiceClient::mock();

        let count_calls = Arc::new(AtomicUsize::new(0));
        let count_calls_clone = count_calls.clone();
        ingester_mock1
            .expect_retain_shards()
            .once()
            .returning(move |request| {
                assert_eq!(request.retain_shards_for_sources.len(), 1);
                assert_eq!(
                    request.retain_shards_for_sources[0].shard_ids,
                    [ShardId::from(1), ShardId::from(3)]
                );
                count_calls_clone.fetch_add(1, Ordering::Release);
                Ok(RetainShardsResponse {})
            });
        ingester_pool.insert("node-1".into(), ingester_mock1.into());
        ingester_pool.insert("node-2".into(), ingester_mock2.into());
        ingester_pool.insert("node-3".into(), ingester_mock3.into());
        let node_id = "node-1".into();
        let wait_handle = ingest_controller.sync_with_ingester(&node_id, &model);
        wait_handle.wait().await;
        assert_eq!(count_calls.load(Ordering::Acquire), 1);
    }

    #[tokio::test]
    async fn test_ingest_controller_advise_reset_shards() {
        let metastore = MetastoreServiceClient::mock().into();
        let ingester_pool = IngesterPool::default();
        let replication_factor = 2;

        let ingest_controller = IngestController::new(metastore, ingester_pool, replication_factor);

        let mut model = ControlPlaneModel::default();

        let index_uid = IndexUid::for_test("test-index", 0);
        let source_id_00: SourceId = "test-source-0".into();
        let source_id_01: SourceId = "test-source-1".into();

        let shards = vec![Shard {
            index_uid: Some(index_uid.clone()),
            source_id: source_id_00.clone(),
            shard_id: Some(ShardId::from(1)),
            shard_state: ShardState::Open as i32,
            publish_position_inclusive: Some(Position::offset(1337u64)),
            ..Default::default()
        }];
        model.insert_shards(&index_uid, &source_id_00, shards);

        let advise_reset_shards_request = AdviseResetShardsRequest {
            shard_ids: vec![
                ShardIds {
                    index_uid: Some(index_uid.clone()),
                    source_id: source_id_00.clone(),
                    shard_ids: vec![ShardId::from(1), ShardId::from(2)],
                },
                ShardIds {
                    index_uid: Some(index_uid.clone()),
                    source_id: source_id_01.clone(),
                    shard_ids: vec![ShardId::from(3)],
                },
            ],
        };
        let advise_reset_shards_response =
            ingest_controller.advise_reset_shards(advise_reset_shards_request, &model);

        assert_eq!(advise_reset_shards_response.shards_to_delete.len(), 2);

        let shard_to_delete_00 = &advise_reset_shards_response.shards_to_delete[0];
        assert_eq!(shard_to_delete_00.index_uid(), &index_uid);
        assert_eq!(shard_to_delete_00.source_id, source_id_00);
        assert_eq!(shard_to_delete_00.shard_ids.len(), 1);
        assert_eq!(shard_to_delete_00.shard_ids[0], ShardId::from(2));

        let shard_to_delete_01 = &advise_reset_shards_response.shards_to_delete[1];
        assert_eq!(shard_to_delete_01.index_uid(), &index_uid);
        assert_eq!(shard_to_delete_01.source_id, source_id_01);
        assert_eq!(shard_to_delete_01.shard_ids.len(), 1);
        assert_eq!(shard_to_delete_01.shard_ids[0], ShardId::from(3));

        assert_eq!(advise_reset_shards_response.shards_to_truncate.len(), 1);

        let shard_to_truncate = &advise_reset_shards_response.shards_to_truncate[0];
        assert_eq!(shard_to_truncate.index_uid(), &index_uid);
        assert_eq!(shard_to_truncate.source_id, source_id_00);
        assert_eq!(shard_to_truncate.shard_positions.len(), 1);
        assert_eq!(
            shard_to_truncate.shard_positions[0].shard_id(),
            ShardId::from(1)
        );
        assert_eq!(
            shard_to_truncate.shard_positions[0].publish_position_inclusive(),
            Position::offset(1337u64)
        );
    }
}
